package org.expeditee.auth;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardCopyOption;
import java.security.InvalidKeyException;
import java.security.KeyFactory;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.SecureRandom;
import java.security.cert.CertificateException;
import java.security.spec.InvalidKeySpecException;
import java.security.spec.PKCS8EncodedKeySpec;
import java.sql.SQLException;
import java.text.ParseException;
import java.util.Base64;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Random;
import java.util.Scanner;
import java.util.stream.Collectors;

import javax.crypto.BadPaddingException;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.NoSuchPaddingException;
import javax.crypto.SecretKey;
import javax.crypto.spec.SecretKeySpec;

import org.expeditee.agents.ExistingFramesetException;
import org.expeditee.agents.InvalidFramesetNameException;
import org.expeditee.auth.Mail.MailEntry;
import org.expeditee.auth.account.Authenticate;
import org.expeditee.auth.account.Authenticate.AuthenticationResult;
import org.expeditee.auth.account.Create;
import org.expeditee.auth.account.Create.CreateResult;
import org.expeditee.auth.account.Password;
import org.expeditee.auth.gui.MailBay;
import org.expeditee.auth.tags.AuthenticationTag;
import org.expeditee.core.Colour;
import org.expeditee.gio.gesture.StandardGestureActions;
import org.expeditee.gui.DisplayController;
import org.expeditee.gui.Frame;
import org.expeditee.gui.FrameIO;
import org.expeditee.gui.MessageBay;
import org.expeditee.items.Item;
import org.expeditee.items.Text;
import org.expeditee.settings.UserSettings;
import org.expeditee.settings.identity.passwordrecovery.Colleagues;
import org.expeditee.settings.identity.secrets.KeyList;
import org.expeditee.stats.Formatter;
import org.ngikm.cryptography.CryptographyConstants;

public class Actions implements CryptographyConstants {
	
	//Debug Functions
	public static void SendTestMessage(String colleagueName) throws InvalidKeySpecException, NoSuchAlgorithmException, FileNotFoundException, KeyStoreException, CertificateException, ClassNotFoundException, IOException, SQLException {
		String time = org.expeditee.stats.Formatter.getDateTime();
		String sender = UserSettings.UserName.get();
		String topic = "Test Message";
		String message = "This is a test message.";
		Map<String, String> options = new HashMap<String, String>();
		options.put("Neat", "Beep");
		MailEntry mail = new MailEntry(time, sender, colleagueName, topic, message, options);
		Mail.sendMail(mail, colleagueName);
		MessageBay.displayMessage("Test message sent.");
	}
	public static void SendTestMessageHemi(String param) {
		String time = Formatter.getDateTime();
		String sender = UserSettings.UserName.get();
		String recipient = param.split(" ")[0];
		String message = param.split(" ")[1];
		Map<String, String> options = new HashMap<String, String>();
		options.put("Accept", "beep");
		options.put("Reject", "beep");
		MailEntry mail = new MailEntry(time, sender, recipient, "Have a key", message, options);
		Mail.sendMail(mail, recipient);
		MessageBay.displayMessage("Test message sent.");
	}
	public static void SendTestOneOffMessage(String colleagueName) {
		String time = Formatter.getDateTime();
		String sender = UserSettings.UserName.get();
		String topic = "Test Message";
		String message = "This is a test message.";
		Map<String, String> options = new HashMap<String, String>();
		options.put("Neat", "Beep");
		MailEntry mail = new MailEntry(time, sender, colleagueName, topic, message, options);
		Random rand = new SecureRandom();
		byte[] key = new byte[16];
		rand.nextBytes(key);
		System.out.println(Base64.getEncoder().encodeToString(key));
		Mail.sendOneOffMail(mail, colleagueName, key);
	}
	public static void AuthOneOffSecureMessage(Text cursor, Text actionItem) {
		byte[] keyBytes = Base64.getDecoder().decode(cursor.getText());
		SecretKey key = new SecretKeySpec(keyBytes, SymmetricAlgorithm);
		List<String> data = actionItem.getData();
		String topic = new String(org.expeditee.auth.sharing.Actions.DecryptSymmetric(Base64.getDecoder().decode(data.get(3)), key));
		String message = new String(org.expeditee.auth.sharing.Actions.DecryptSymmetric(Base64.getDecoder().decode(data.get(4)), key));
		Map<String, String> options = new HashMap<String, String>();
		for (int i = 5; i < data.size(); i+=2) {
			String k = new String(org.expeditee.auth.sharing.Actions.DecryptSymmetric(Base64.getDecoder().decode(data.get(i)), key));
			String v = new String(org.expeditee.auth.sharing.Actions.DecryptSymmetric(Base64.getDecoder().decode(data.get(i + 1)), key));
			options.put(k, v);
		}
		MailBay.addMessage(data.get(0), topic, message, options);
		StandardGestureActions.Refresh();
	}
	public static void SetPWColleagues(String colleagueOne, String colleagueTwo) {
		Colleagues.Colleague_One.set(colleagueOne);
		Colleagues.Colleague_Two.set(colleagueTwo);
		Colleagues.Colleague_One_Email.set("bnemhauser@gmail.com");
		Colleagues.Colleague_Two_Email.set("bnemhauser@gmail.com");
	}
	
	private static String userbackup = "authadmin";
	public static void ToggleAuth() {
		String backup = UserSettings.UserName.get();
		UserSettings.UserName.set(userbackup);
		userbackup = backup;
	}
	
	/**
	 * Display Expeditee Mail
	 * @throws IOException 
	 * @throws SQLException 
	 * @throws ClassNotFoundException 
	 * @throws CertificateException 
	 * @throws NoSuchAlgorithmException 
	 * @throws FileNotFoundException 
	 * @throws KeyStoreException 
	 * @throws ParseException 
	 * @throws InvalidKeySpecException 
	 * @throws BadPaddingException 
	 * @throws IllegalBlockSizeException 
	 * @throws NoSuchPaddingException 
	 * @throws InvalidKeyException 
	 */
	public static void ToggleBay() throws KeyStoreException, FileNotFoundException, NoSuchAlgorithmException, CertificateException, ClassNotFoundException, SQLException, IOException, ParseException, InvalidKeySpecException, InvalidKeyException, NoSuchPaddingException, IllegalBlockSizeException, BadPaddingException {
		if (!AuthenticatorBrowser.isAuthenticated()) return;
		if (!DisplayController.isMailMode()) {
			MailBay.ensureLink();
			Mail.clear();
			String keyEncoded = KeyList.PrivateKey.get().getData().get(0);
			byte[] keyBytes = Base64.getDecoder().decode(keyEncoded);
			PrivateKey key = KeyFactory.getInstance(AsymmetricAlgorithm).generatePrivate(new PKCS8EncodedKeySpec(keyBytes));
			Mail.checkMail(key);
		}
		DisplayController.ToggleMailMode();
	}
 	
	/**
	 * Action used to navigate the authorised user back to their desktop.
	 */
	public static void AuthGoToDesktop() {
		DisplayController.setCurrentFrame(FrameIO.LoadFrame(UserSettings.HomeFrame.get()), true);
	}
	
	/**
	 * Action used to created a new user account.
	 * Attempts to use content from text items on frame, will default to java properties if they cannot be found.
	 * Will fail if it cannot find content from text items on frame and all required java properties are not present.
	 * @throws SQLException 
	 * @throws IOException 
	 * @throws ExistingFramesetException 
	 * @throws InvalidFramesetNameException 
	 * @throws ClassNotFoundException 
	 * @throws FileNotFoundException 
	 * @throws CertificateException 
	 * @throws NoSuchAlgorithmException 
	 * @throws KeyStoreException 
	 * @throws BadPaddingException 
	 * @throws IllegalBlockSizeException 
	 * @throws NoSuchPaddingException 
	 * @throws InvalidKeySpecException 
	 * @throws InvalidKeyException 
	 * @throws ParseException 
	 * @throws Exception 
	 */
	public static void AuthCreateAccount() throws KeyStoreException, NoSuchAlgorithmException, CertificateException, FileNotFoundException, ClassNotFoundException, InvalidFramesetNameException, ExistingFramesetException, IOException, SQLException, InvalidKeyException, InvalidKeySpecException, NoSuchPaddingException, IllegalBlockSizeException, BadPaddingException, ParseException {
		Collection<Text> textItems = DisplayController.getCurrentFrame().getTextItems();
		Optional<Map<AuthenticationTag, String>> userdata = 
				AuthenticationTag.fetchUserData(textItems, false, 
						AuthenticationTag.Username,
						AuthenticationTag.Password,
						AuthenticationTag.PasswordAgain, 
						AuthenticationTag.Email, 
						AuthenticationTag.EmailAgain);
		
		if (userdata.isPresent()) {
			Map<AuthenticationTag, String> userData = userdata.get();
			
			// A profile already existing with 'username' means an account cannot be created with that username.			
			if (FrameIO.getProfilesList().contains(userData.get(AuthenticationTag.Username))) {
				MessageBay.errorMessage("A Expeditee profile with this username already exists, please choose another.");
				return;
			}
			
			// The chosen username must be a valid frameset name.
			if (!FrameIO.isValidFramesetName(userData.get(AuthenticationTag.Username))) {
				MessageBay.errorMessage("The provided username must begin and end with a letter and contain only letters and numbers inbetween, please choose another.");
				return;
			}
			
			// The passwords provided must match
			if (userData.get(AuthenticationTag.Password).compareTo(userData.get(AuthenticationTag.PasswordAgain)) != 0) {
				MessageBay.errorMessage("The provided passwords do not match, please fix this and try again.");
				return;
			}
			
			// The emails provided must match
			if (userData.get(AuthenticationTag.Email).compareTo(userData.get(AuthenticationTag.EmailAgain)) != 0) {
				MessageBay.errorMessage("The provided emails do not match, please fix this and try again.");
				return;
			}
			
			CreateResult result = Create.createAccount(userData);
			if (result == CreateResult.SuccessCreateAccount) {
				Authenticate.login(userData);
				AuthenticatorBrowser.Authenticated = true;
			} else {
				MessageBay.errorMessage(result.toString());
			}
		} else {
			MessageBay.errorMessage("Please fill out all the supplied text boxes.");
		}
	}
	
	/**
	 * Action used to start authentication as a specified user.  
	 * Attempts to use content from text items on frame, will default to java properties if they cannot be found.
	 * Will fail if it cannot find content from text items on frame and all required java properties are not present.
	 * @throws Exception 
	 */
	public static void AuthLogin() {
		Collection<Text> textItems = DisplayController.getCurrentFrame().getTextItems();
		Optional<Map<AuthenticationTag, String>> userdata = AuthenticationTag.fetchUserData(textItems, false, AuthenticationTag.Username, AuthenticationTag.Password);
		if (userdata.isPresent()) {
			AuthenticationResult result = Authenticate.login(userdata.get());
			if (result == AuthenticationResult.SuccessLogin) {
				MessageBay.displayMessage(result.toString());
				AuthenticatorBrowser.Authenticated = true;
			} else {
				MessageBay.errorMessage(result.toString());
			}
		} else {
			MessageBay.errorMessage("Please fill out all the supplied text boxes.");
		}
	}
	
	public static void AuthLogout() {
		MessageBay.displayMessage(Authenticate.logout().toString());
	}
	
	/**
	 * Action used to change the currently authenticated users password.
	 * Attempts to use content from text items on frame, will default to java properties if they cannot be found.
	 * Will fail if it cannot find content from text items on frame and all required java properties are not present.
	 * Will fail if no user is currently logged in.
	 * @throws IOException 
	 * @throws CertificateException 
	 * @throws FileNotFoundException 
	 * @throws KeyStoreException 
	 * @throws NoSuchAlgorithmException 
	 * @throws SQLException 
	 * @throws ClassNotFoundException 
	 */
	public static void AuthChangePassword() throws NoSuchAlgorithmException, KeyStoreException, FileNotFoundException, CertificateException, IOException, ClassNotFoundException, SQLException {
		final Collection<Text> textItems = DisplayController.getCurrentFrame().getTextItems();
		
		if (!AuthenticatorBrowser.Authenticated) {
			MessageBay.errorMessage("You must be logged in to perform this action.");
		} else {
			final Optional<Map<AuthenticationTag, String>> userdata = AuthenticationTag.fetchUserData(textItems, false, AuthenticationTag.Password, AuthenticationTag.NewPassword, AuthenticationTag.NewPasswordAgain);
			if (userdata.isPresent()) {
				final Map<AuthenticationTag, String> userData = userdata.get();
				if (userData.get(AuthenticationTag.NewPassword).compareTo(userData.get(AuthenticationTag.NewPasswordAgain)) != 0) {
					MessageBay.errorMessage("The provided passwords do not match, please fix this and try again.");
				} else {
					userData.put(AuthenticationTag.Username, UserSettings.UserName.get());
					Password.changePassword(userData);
				}
			} else {
				MessageBay.errorMessage("Please fill out all the supplied text boxes.");
			}
		}
	}
	
	public static void AuthGotoAccountManagement() {
		if (AuthenticatorBrowser.Authenticated) {
			DisplayController.setCurrentFrame(FrameIO.LoadFrame("multiuser1"), false);
		} else {
			DisplayController.setCurrentFrame(FrameIO.LoadFrame("authentication1"), false);
		}
	}
	
	public static void AuthDistributeIntergalacticNumber() {
		Collection<Text> textItems = DisplayController.getCurrentFrame().getTextItems();
		Optional<Map<AuthenticationTag, String>> userdata = 
				AuthenticationTag.fetchUserData(textItems, false, AuthenticationTag.Username);
		if (userdata.isPresent()) {
			Map<AuthenticationTag, String> userData = userdata.get();
			String username = userData.get(AuthenticationTag.Username);
			String email = getEmailFromUsername(username);
			userData.put(AuthenticationTag.Email, email);
			Password.generateAndDeliverIntergalacticNumber(userData);
			MessageBay.displayMessage("A identity number has been sent to the email "
					+ "associated with your account.  Enter it below to proceed.");
		}
	}
	
	public static void AuthSubmitIntergalacticNumber() {
		Collection<Text> textItems = DisplayController.getCurrentFrame().getTextItems();
		Optional<Map<AuthenticationTag, String>> userdata = 
				AuthenticationTag.fetchUserData(textItems, false, AuthenticationTag.Username, 
						AuthenticationTag.IntergalacticNumber);
		if (userdata.isPresent()) {
			Map<AuthenticationTag, String> tags = userdata.get();
			String username = tags.get(AuthenticationTag.Username);
			String intergalacticNumber = tags.get(AuthenticationTag.IntergalacticNumber);
			boolean match = false;
			try {
				match = AuthenticatorBrowser.getInstance().confirmIntergalaticNumber(username, intergalacticNumber);
			} catch (NoSuchAlgorithmException | KeyStoreException | CertificateException | ClassNotFoundException
					| IOException | SQLException e) {
				e.printStackTrace();
				return;
			}
			if (!match) {
				MessageBay.errorMessage("The provided identity number does not match the one stored on file.");
				return;
			}
			String[] colleagues = getPasswordColleaguesFromUsername(username);
			Password.confirmIntergalacticNumberAndAlertColleagues(userdata.get(), colleagues);
		}
	}
	
	private static String getEmailFromUsername(String username) {
		Path credentialsDirPath = Paths.get(FrameIO.PROFILE_PATH).resolve(username).resolve(username + "-credentials");
		Path credentialsFilePath = credentialsDirPath.resolve("credentials.inf");
		String fileName = null;
		if (credentialsFilePath.toFile().exists()) {
			try (Scanner in = new Scanner(credentialsFilePath)) {
				fileName = in.nextLine();
			} catch (IOException e) {
				MessageBay.errorMessage("Unable to locate public email for specified user, are they registered on this computer?");
				return null;
			}
		} else {
			MessageBay.errorMessage("Unable to locate public email for specified user, are they registered on this computer?");
			return null;
		}
		
		int number = Integer.parseInt(fileName.replace(".exp", ""));
		Frame credentialsFrame = FrameIO.LoadFrame(username + number, FrameIO.PROFILE_PATH);
		Collection<Text> textItems = credentialsFrame.getTextItems();
		textItems.removeIf(text -> !text.getText().startsWith("Email: "));
		if (textItems.isEmpty()) {
			MessageBay.errorMessage("Unable to locate public email for specified user, are they registered on this computer?");
			return null;
		} else {
			Text emailText = textItems.iterator().next();
			String email = emailText.getText().replace("Email: ", "");
			return email;
		}
	}
	
	private static String[] getPasswordColleaguesFromUsername(String username) {
		Path credentialsDirPath = Paths.get(FrameIO.PROFILE_PATH).resolve(username).resolve(username + "-credentials");
		Path credentialsFilePath = credentialsDirPath.resolve("pwcolleagues.inf");
		String fileName = null;
		if (credentialsFilePath.toFile().exists()) {
			try (Scanner in = new Scanner(credentialsFilePath)) {
				fileName = in.nextLine();
			} catch (IOException e) {
				MessageBay.errorMessage("Unable to password colleague frame for specified user, are they registered on this computer?");
				return null;
			}
		} else {
			MessageBay.errorMessage("Unable to password colleague frame for specified user, are they registered on this computer?");
			return null;
		}
		
		int number = Integer.parseInt(fileName.replace(".exp", ""));
		Frame pwColleagueFrame = FrameIO.LoadFrame(username + number, FrameIO.PROFILE_PATH);
		Collection<Text> textItems = pwColleagueFrame.getTextItems();
		textItems.removeIf(text -> !text.getText().startsWith("Colleague"));
		
		String[] ret = new String[4];
		Iterator<Text> it = textItems.iterator();
		while(it.hasNext()) {
			String content = it.next().getText().toLowerCase().trim();
			if (content.contains("colleague_one:")) {
				ret[0] = content.replace("colleague_one:", "").trim();
			} else if (content.contains("colleague_two:")) {
				ret[1] = content.replace("colleague_two:", "").trim();
			} else if (content.contains("colleague_one_email:")) {
				ret[2] = content.replace("colleague_one_email:", "").trim();
			} else if (content.contains("colleague_two_email:")) {
				ret[3] = content.replace("colleague_two_email:", "").trim();
			}
		}
		return ret;
	}
	
	public static void AuthShareFrameset() throws IOException {
		Collection<Text> textItems = DisplayController.getCurrentFrame().getTextItems();
		
		Optional<Map<AuthenticationTag, String>> userdata = AuthenticationTag.fetchUserData(textItems, false, AuthenticationTag.Frameset);
		if (userdata.isPresent()) {
			Map<AuthenticationTag, String> userData = userdata.get();
			FrameIO.SuspendCache();
			Frame toShare = FrameIO.LoadFrame(userData.get(AuthenticationTag.Frameset) + 1);
			FrameIO.ResumeCache();
			
			if (toShare == null) {
				MessageBay.errorMessage("Insufficient information provided to complete this action.");
				return;
			}
			
			shareFrameset(toShare);
		}
	}

	/**
	 * Navigation action for progressing the process of recruiting colleagues to assist in password recovery.
	 * Hides certain content that AuthSubmitPWCollegues goes onto show if it does not fail.
	 */
	public static void AuthGotoColleagueSubmissionFrame() {
		Frame destination = FrameIO.LoadFrame("authentication7");
		DisplayController.setCurrentFrame(destination, true);
		Collection<Item> toHide = getByData(destination, "ShowOnProgress");
		for (Item i: toHide) {
			i.setVisible(false);
		}
	}
	
	/**
	 * Action used to start the process of formalising the password recovery process.
	 * @throws SQLException 
	 * @throws IOException 
	 * @throws ClassNotFoundException 
	 * @throws CertificateException 
	 * @throws NoSuchAlgorithmException 
	 * @throws FileNotFoundException 
	 * @throws KeyStoreException 
	 * @throws InvalidKeySpecException 
	 */
	public static void AuthSubmitPWColleagues() throws InvalidKeySpecException, NoSuchAlgorithmException, KeyStoreException, FileNotFoundException, CertificateException, ClassNotFoundException, IOException, SQLException {
		Frame currentFrame = DisplayController.getCurrentFrame();
		Collection<Text> textItems = currentFrame.getTextItems();
		
		if (!AuthenticatorBrowser.Authenticated) {
			MessageBay.errorMessage("You must be logged in to perform this action.");
			return;
		}
		
		Optional<Map<AuthenticationTag, String>> userdata = AuthenticationTag.fetchUserData(textItems, false, AuthenticationTag.ColleagueOne, AuthenticationTag.ColleagueTwo);
		if (userdata.isPresent()) {
			Map<AuthenticationTag, String> userData = userdata.get();
			String colleagueOne = userData.get(AuthenticationTag.ColleagueOne);
			Path colleagueOnePath = Paths.get(FrameIO.CONTACTS_PATH).resolve(colleagueOne + "-credentials");
			String colleagueTwo = userData.get(AuthenticationTag.ColleagueTwo);
			Path colleagueTwoPath = Paths.get(FrameIO.CONTACTS_PATH).resolve(colleagueTwo + "-credentials");
			if (!colleagueOnePath.toFile().exists()) {
				MessageBay.errorMessage("Your nominated colleague: " + colleagueOne + " must exist in your contacts.");
			} else if (!colleagueTwoPath.toFile().exists()) {
				MessageBay.errorMessage("Your nominated colleague: " + colleagueTwo + " must exist in your contacts.");
			} else {
				userData.put(AuthenticationTag.Username, UserSettings.UserName.get());
				boolean success = submitPWColleagues(userData);
				if (success) {
					Collection<Item> toShow = getByData(currentFrame, "ShowOnProgress");
					for (Item i: toShow) {
						i.setVisible(true);
					}
					currentFrame.change();
					MessageBay.displayMessage("-------Messages sent-------");
				}
				FrameIO.SaveFrame(currentFrame);
				DisplayController.requestRefresh(false);
			}
		}
	}
	
//	public static void AuthSetupPasswordRecovery() throws KeyStoreException, FileNotFoundException, NoSuchAlgorithmException, CertificateException, ClassNotFoundException, IOException, SQLException, UnrecoverableEntryException {
//		if (!UserSettings.Authenticated.get()) {
//			MessageBay.errorMessage("You must be logged in to perform this action.");
//		} else if (!Authenticator.getInstance().hasRegisteredEmail(UserSettings.UserName.get())) {
//			Frame registerEmailFrame = FrameIO.LoadFrame("authentication4");
//			DisplayController.setCurrentFrame(registerEmailFrame, true);
//		} else if (!Authenticator.getInstance().hasRequestedColleagues(UserSettings.UserName.get()) && Authenticator.getInstance().getColleagues(UserSettings.UserName.get()) == null) {
//			Frame submitColleaguesFrame = FrameIO.LoadFrame("authentication5");
//			DisplayController.setCurrentFrame(submitColleaguesFrame, true);
//		} else if (Authenticator.getInstance().hasRequestedColleagues(UserSettings.UserName.get()) && Authenticator.getInstance().getColleagues(UserSettings.UserName.get()) == null) {
//			MessageBay.displayMessage("You have already nominated two colleagues to assist you in the process of password recovery and are awaiting their response."
//					+ "  You will be alerted on Expeditee startup when they have both responded.");
//		} else if (Authenticator.getInstance().getColleagues(UserSettings.UserName.get()) != null) {
//			MessageBay.displayMessage("You have completed the Password Recovery Setup process, there is nothing more to do here.");
//		}
//	}
	
	public static void AuthConfirmPasswordColleagueRelationship(String colleagueName) {
		
	}
	
	public static void AuthDenyPasswordColleagueRelationship(String colleagueName) throws InvalidKeySpecException, NoSuchAlgorithmException, 
		KeyStoreException, FileNotFoundException, CertificateException, ClassNotFoundException, IOException, SQLException {
		denyPasswordColleagueRelationship(colleagueName);
	}
	
	public static void AuthClearPWColleaguesNominated() {
		
	}
	
	/*
	 * Function to share a specified frameset.
	 * Currently, this moves the frameset to the 'Shared By Me' directory and then relies on the user to use Google Drive functionality to share it appropriately.
	 */
	private static void shareFrameset(Frame toShare) throws IOException {	
		File destinationDir = new File(FrameIO.SHARED_FRAMESETS_PATH + File.separator + toShare.getFramesetName());
		File sourceDir = new File(toShare.getFramesetPath());
		
		if (destinationDir.exists()) {
			MessageBay.errorMessage("A frameset by this name already exists.");
			return;
		}
		
		destinationDir.mkdir();
		List<Path> files = Files.walk(sourceDir.toPath()).collect(Collectors.toList());
		Files.move(files.get(0), destinationDir.toPath(), StandardCopyOption.ATOMIC_MOVE);
		
		MessageBay.displayMessage("The frameset " + toShare.getFramesetName() + " has been moved to " + destinationDir + ".  Google Drive functionality can now be used to share it with colleagues.");
	}
	
	private static void denyPasswordColleagueRelationship(String colleagueName) throws InvalidKeySpecException, NoSuchAlgorithmException, KeyStoreException, FileNotFoundException, CertificateException, ClassNotFoundException, IOException, SQLException {
		String time = org.expeditee.stats.Formatter.getDateTime();
		String sender = UserSettings.UserName.get();
		String message = "You have received a reply from " + sender + " reguarding your request for assistance.";
		String message2 = "Unfortunately " + sender + " has indicated that they are unable to help you with your potential password recovery.";
		Map<String, String> options = new HashMap<String, String>();
		options.put("Clear Preview Colleague Nominations", "AuthClearPWColleaguesNominated");
		MailEntry mail = new MailEntry(time, sender, colleagueName, message, message2, options);
		Mail.sendMail(mail, colleagueName);
	}
	
	private static boolean submitPWColleagues(Map<AuthenticationTag, String> userData) throws InvalidKeySpecException, NoSuchAlgorithmException, KeyStoreException, FileNotFoundException, CertificateException, ClassNotFoundException, IOException, SQLException {
		String colleagueOne = userData.get(AuthenticationTag.ColleagueOne);
		String colleagueTwo = userData.get(AuthenticationTag.ColleagueTwo);
		PublicKey colleagueOneKey = AuthenticatorBrowser.getInstance().getPublicKey(colleagueOne);
		PublicKey colleagueTwoKey = AuthenticatorBrowser.getInstance().getPublicKey(colleagueTwo);
		if (colleagueOneKey == null) {
			MessageBay.errorMessage("Unable to get public key for colleague: " + colleagueOne);
			return false;
		} else if (colleagueTwoKey == null) {
			MessageBay.errorMessage("Unable to get public key for colleague: " + colleagueTwo);
			return false;
		} else {
			String time = org.expeditee.stats.Formatter.getDateTime();
			String sender = userData.get(AuthenticationTag.Username);
			String topic = "You have received a request for cooperation from your colleague " + sender;
			String message = "Should " + sender + " forget their password, they would like your help recoverying it.";
			Map<String, String> arguments = new HashMap<String, String>();
			arguments.put("I agree to assist " + sender + " if they loose access to their account.", "AuthConfirmPasswordColleagueRelationship " + sender);
			arguments.put("I wish to excuse myself from this responsibility.", "AuthDenyPasswordColleagueRelationship " + sender);
			MailEntry mail = new MailEntry(time, sender, colleagueOne, topic, message, arguments);
			Mail.sendMail(mail, colleagueOne);
			mail = new MailEntry(time, sender, colleagueTwo, topic, message, arguments);
			Mail.sendMail(mail, colleagueTwo);
			AuthenticatorBrowser.getInstance().markRequestedColleagues(UserSettings.UserName.get());
			return true;
		}
	}
	
	public static void TickBox(final Text item) {
		if (item.getBackgroundColor() != Colour.RED) {
			item.setBackgroundColor(Colour.RED);
		} else {
			item.setBackgroundColor(Colour.GREEN);
		}
	}
	
	/*
	 * Gets all items on a specified frame that contain the specified data.
	 */
	public static Collection<Item> getByData(final Frame frame, final String data) {
		final Collection<Item> allItems = frame.getAllItems();
		allItems.removeIf(i -> i.getData() == null || !i.hasData(data));
		return allItems;
	}
	
	public static Collection<Item> getByContent(final Frame frame, final String content) {
		final Collection<Item> allItems = frame.getAllItems();
		allItems.removeIf(i -> i.getText().compareTo(content) != 0);
		return allItems;
	}
}
